// Copyright 2023 Sirbu Dan
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//     http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <io.hpp>
#include <serial.hpp>
#include <timer.hpp>
#include <interrupt.hpp>
#include <address.hpp>
#include <state.hpp>
#include <rfid.hpp>
#include <pins.hpp>
#include <LiquidCrystal_I2C.h>
#include <Keypad.h>
#include <EEPROM.h>
#include <MFRC522.h>

LiquidCrystal_I2C lcd(0x27, 16, 2);
MFRC522 mfrc522(SS_PIN, RST_PIN);
MFRC522::MIFARE_Key rfid_key;
void(* resetFunc) (void) = 0;

const byte ROWS = 4; 
const byte COLS = 3; 

char hexaKeys[ROWS][COLS] = {
  {'1', '2', '3'},
  {'4', '5', '6'},
  {'7', '8', '9'},
  {'*', '0', '#'}
};

byte rowPins[ROWS] = {8, 7, 6, 5}; 
byte colPins[COLS] = {A3, A0, A2}; 

Keypad customKeypad = Keypad(makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS);

String def = "Pressed: ";
String passPrompt = "PASS: ";
String pass;
char id[9];
int state;
int passLen;
int cardWrote;
byte rfid_block[16] = {"PM_LOK\0\0\0\0\0\0\0\0"};
byte rfid_read[18];

void buttonInterrupt(void* arg)
{
  write(RELAY_PIN, HIGH);
  write(LED_PIN, LOW);
  tone(BUZZER_PIN, BUZZER_TONE);
  setTimerRunningNoCLI(true);
}

void unlock()
{
  write(RELAY_PIN, HIGH);
  write(LED_PIN, LOW);
  tone(BUZZER_PIN, BUZZER_TONE);
  setTimerRunning(true);
}

void setup() {
  passLen = 0;
  cardWrote = 0;

  for (int i = 0; i < 6; i++) {
    rfid_key.keyByte[i] = 0xFF;
  }

  setPinMode(BUTTON_PIN, INPUT_PULLUP);
  setPinMode(RELAY_PIN, OUTPUT);
  setPinMode(BUZZER_PIN, OUTPUT);
  setPinMode(LED_PIN, OUTPUT);

  write(LED_PIN, HIGH);

  serialBegin();
  timerSetup();
  setupInterrupts();
  setButtonCallback(buttonInterrupt, NULL);

  SPI.begin();
  mfrc522.PCD_Init();

  lcd.init();       
  lcd.backlight();
  lcd.clear();
  lcd.setCursor(0,0);

  EEPROM.get(EEPROM_ID_ADDRESS, id);

  if (strncmp("LOCKER_PM", id, 9) != 0) {
    lcd.print(passPrompt.c_str());
    state = STATE_INIT_PASS;
  } else {
    lcd.print(passPrompt.c_str());
    state = STATE_NORMAL;
  }
}

void loop() {
  if (state == STATE_INIT_PASS) {
    if (passLen == PASS_LEN) {
      lcd.clear();
      lcd.setCursor(0,0);
      lcd.print("Aprroach the");
      lcd.setCursor(0, 1);
      lcd.print("rfid tag");
      state = STATE_INIT_TAG;
    } else {
      char passKey = customKeypad.getKey();

      if (passKey) {
        passLen++;
        pass += passKey;

        String print = passPrompt;
        print += pass;

        lcd.clear();
        lcd.setCursor(0,0);
        lcd.print(print.c_str());
      }
    }
  } else if (state == STATE_INIT_TAG) {
    if (!cardWrote) {
      if (mfrc522.PICC_IsNewCardPresent()) {
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.print("Hold the tag");

        if (mfrc522.PICC_ReadCardSerial()) {
          writeBlock(mfrc522, rfid_key, RFID_BLOCK, rfid_block);
          cardWrote = 1;
        }
      }
    } else {
      EEPROM.put(EEPROM_ID_ADDRESS, "LOCKER_PM");
      char passBuff[8];
      strncpy(passBuff, pass.c_str(), 8);
      EEPROM.put(EEPROM_PASS_ADDRESS, passBuff);

      resetFunc();
    }
  } else {
    if (mfrc522.PICC_IsNewCardPresent()) {
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.print("Hold the tag");

        if (mfrc522.PICC_ReadCardSerial()) {
          readBlock(mfrc522, rfid_key, RFID_BLOCK, rfid_read);
        }

        if (strcmp((char*)rfid_read, "PM_LOK") == 0) {
          lcd.clear();
          lcd.setCursor(0,0);
          lcd.print(passPrompt.c_str());
          unlock();
          pass = "";
          passLen = 0;
          memset(rfid_read, 0, 18);
        }

        mfrc522.PCD_Reset();
        mfrc522.PCD_Init();
      } else {
        if (passLen == PASS_LEN) {
          char passBuff[8];
          EEPROM.get(EEPROM_PASS_ADDRESS, passBuff);

          if (strncmp(passBuff, pass.c_str(), 8) == 0) {
            lcd.clear();
            lcd.setCursor(0,0);
            lcd.print(passPrompt.c_str());
            unlock();
            pass = "";
            passLen = 0;
          } else if (strncmp(pass.c_str(), "*#0000*#", 8) == 0) {
            EEPROM.put(EEPROM_ID_ADDRESS, "UPDATE_ME");
            resetFunc();
          } else {
            pass = "";
            passLen = 0;
            lcd.clear();
            lcd.setCursor(0,0);
            lcd.print(passPrompt.c_str());
          }
        } else {
          char passKey = customKeypad.getKey();

          if (passKey) {
            passLen++;
            pass += passKey;

            String print = passPrompt;
            print += pass;

            lcd.clear();
            lcd.setCursor(0,0);
            lcd.print(print.c_str());
          }
        }
      }
  }
}